''
'' UCI interface for Saitek-OSA-Interace (COM port)
'' Optimized for Arena
'' V1.2
''
'' (C) 2009 by Achim Pietig <achim@pietig.com>
'' License:
''
'' Redistribution and use in source and binary forms, with or without
'' modification, are permitted provided that the following conditions
'' are met:
''
'' 1. Redistributions of source code must retain the above copyright
''    notice, this list of conditions and the following disclaimer.
'' 2. Redistributions in binary form must reproduce the above copyright
''    notice, this list of conditions and the following disclaimer in the
''    documentation and/or other materials provided with the distribution.
'' 3. The name of the author may not be used to endorse or promote products
''    derived from this software without specific prior written permission.
''
'' Changes to this license can be made only by the copyright author with
'' explicit written consent.
''
'' THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
'' IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
'' OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
'' IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
'' INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
'' NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
'' DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
'' THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
'' (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
'' THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

' Lib for Function FileExists

#include "file.bi"

' Setup board position from FEN string
Declare Sub setuppos(ByRef fenstr As String)

'
'   Main program
'

DIM As UShort n, i
DIM As String indata, dat, cdata, oldFEN, bestmove, actmove, pv
DIM As String actCOM, actBaud, oldpv, bname, bmodul, lev
Dim Shared nParams As UByte
ReDim Shared Param(1 To 2) As String
' Sets default command line parameters
actCOM = "COM1:"
actBaud = "1200"

' Checking for INI file
If FileExists("osa.ini") Then
    Open "osa.ini" for Input As #4
    Do Until EOF(4)
        Line Input #4, indata
        indata = Trim(indata) : dat = ""
        ' Strip off blanks
        For i = 1 to Len(indata)
            If Mid(indata, i, 1) <> " " Then dat = dat + Mid(indata, i, 1)
        Next i
        If LCase(Left(dat, 4)) = "com=" Then actCOM = Mid(dat, 5)
        If LCase(Left(dat, 5)) = "baud=" Then actBaud = Mid(dat, 6)
        If LCase(Left(dat, 5)) = "name=" Then bname = Mid(dat, 6)
        If LCase(Left(dat, 6)) = "modul=" Then bmodul = Mid(dat, 7)
        If LCase(Left(dat, 6)) = "level=" Then lev = Mid(dat, 7)
    Loop
    Close #4                        
End If

' Checking for command line parameters
nParams = 0
While Command(nParams + 1) <> ""
    nParams = nParams + 1
Wend
If nParams > 0 Then
    ReDim Param(1 To nParams) As String
    For i = 1 To nParams
        Param(i) = Command(i)
    Next i
End If

' Help
If UCase(Param(1)) = "/?" Or UCase(Param(1)) = "/H" Then
    Print
    Print "OSA to UCI converter for Scisys/Saitek chess boards"
    Print "(C) 2009 by Achim Pietig, Version 1.2"
    Print
    Print "Usage: OSA2UCI [/?] Or [/h]"
    Print "       OSA2UCI [COMx:] [baud]"
    Print "Call without parameters sets COM1: with 1200 baud"
    Print "Board shall be set to corresponding baud rate before!"
    Print
    Print "e.g.   OSA2UCI                (COM1: with 1200 baud)"
    Print "       OSA2UCI COM2:          (COM2: with 1200 baud)"
    Print "       OSA2UCI COM3: 9600     (COM3: with 9600 baud)"
    End
End If

' Init program parameters
If nParams > 0 Then actCOM = UCase(Param(1))
If nParams > 1 Then actBaud = Param(2)

' Opens related COM port with board specific values
' OSA cable supports TxD, RxD only. All hardware specific handshakes are disabled.
Open Cons For Input As #1
If Open Com (actCOM + actbaud + ",N,8,1,CS0,DS0,CD0,RS,BIN" For Binary As #2) <> 0 Then
    Print "Cannot initialise comport"
    Close
    End
End If

'debug
'Open "OSA_log.txt" For Output As #3

' Sets OSA to english language for all following commands
Print #2, "ENGLISH"
' There is no response, so wait 250 ms to give the board some time before sending the next command
Sleep 250,1

Print #2, "OPEN"
Do
    dat = Input(1, #2)
    If Asc(dat) > 31 And dat <> ">" Then cdata = cdata + dat
Loop Until dat = Chr(&H17)
' Open command returns an initial string from the board with OSA version and actual baud rate
' The answer is shown as info string in Arena (engine debug window)
' Correct answers of the board end with a character code of &H17, so most commands of this program will receive data until Chr$(&H17) is detected.
Print "info string "; Trim(cdata)

' Switches board off, pieces and LEDs on the board will not be used
Print #2, "BOARD OFF"
Do
    dat = Input(1, #2)
Loop Until dat = Chr(&H17)

' Switch off automatic responses of the board during calculation
' Board should only send answers to commands from this program
Print #2, "SENDINFO OFF"
Do
    dat = Input(1, #2)
Loop Until dat = Chr(&H17)

' Sets initial game level, if available
If Len(lev) = 2 Then
    Print #2, "SETLEVEL "; lev
    Do
        dat = Input(1, #2)
    Loop Until dat = Chr(&H17)
End If

' Program loop
Do
        ' Waits for an Arena UCI command
        Line Input #1, indata
        indata = Trim(indata)
        cdata = "" : dat = ""

        ' Flush COM buffer
        ' Clears the buffer from unnecessary bytes
        While LOC(2) > 0
            dat = Input(LOC(2), #2)
        Wend

        ' Answer to Arena in case of 'uci'
        If indata = "uci" Then
            If bname + bmodul = "" Then
                Print "id name OSA2UCI V1.2"
                Print "id author Achim Pietig"
            Else
                Print "id name "; bname;
                If bmodul <> "" Then Print " with "; bmodul;
                Print
            End If
            ' Typical values for chess computers with own book
            Print "option name OwnBook type check default true"
            ' Set up start position
            Print #2, "NEWGAME"
            Do
                dat = Input(1, #2)
                If Asc(dat) > 31 And dat <> ">" Then cdata = cdata + dat
            Loop Until dat = Chr(&H17)
            oldpv = ""
            If Instr(cdata, "New Game") Then Print "uciok"
        End If

        ' Answer to Arena in case of 'isready'
        If indata = "isready" Then
            Print "readyok"
        End If

        ' Action in case of 'ucinewgame', no response to Arena
        If indata = "ucinewgame" Then
            ' Set up start position
            Print #2, "NEWGAME"
            Do
                dat = Input(1, #2)
            Loop Until dat = Chr(&H17)
            oldPV = ""
        End If

        ' Specific command 'board', returns current position in OSA notation
        If indata = "board" Then
            Print #2, "POSITION"
            Do
                dat = Input(1, #2)
                If Asc(dat) > 31 And dat <> ">" Then cdata = cdata + dat
                If Asc(dat) = 13 Then cdata = cdata + dat
                If Asc(dat) = 10 Then cdata = cdata + dat
            Loop Until dat = Chr(&H17)
            Print cdata
        End If

        ' Specific command 'setlevel', sets a new level in the board and displays the result
        If Left(indata, 8) = "setlevel" Then
            Print #2, indata
            Do
                dat = Input(1, #2)
                If Asc(dat) > 31 And dat <> ">" Then cdata = cdata + dat
            Loop Until dat = Chr(&H17)
            Print cdata
        End If

        ' Action in case of 'position' command from Arena, no response
        ' Arena sends a complete move line, only the last move is needed for the board and stored in actmove
        ' If a fen string is sent, the setuppos function is called and converts the fen string to OSA position commands
        If Left(indata, 9) = "position " Then
            n = Instr(10, indata, "moves ") : actmove = ""
            If n > 0 Then
                ' Detects a promotion
                If Instr(1, "qrnb", Right(indata, 1)) > 0 Then
                    actmove = Right(indata, 5)
                Else
                    actmove = Right(indata, 4)
                End If
                ' Move notation from Arena is converted to OSA
                actmove = Left(actmove, 2) + "-" + Mid(actmove, 3)
                ' A promotion piece is added
                If Instr(1, "qrnb", Right(actmove, 1)) > 0 Then actmove = Left(actmove, 5) + "/" + UCase(Right(actmove, 1))
                indata = Left(indata, n - 2)
            End If
            n = Instr(10, indata, "fen ")
            If n > 0 Then
                cdata = Trim(Mid(indata, n + 4))
                ' A fen string is converted and send to the board only, if it differs from a previous one (normally one time per game)
                ' This is needed, because Arena sends a startup position or fen string for every move
                If cdata <> oldFEN Then
                    oldFEN = cdata
                    ' change board position
                    setuppos(cdata)
                End If
            End If
        End If

        ' Action in case of 'go' command from Arena, response is 'bestmove' with the move from the board
        ' If actmove is empty (first command for white), then a PLAY command is sent to the board, otherwise a MOVE command
        If Left(indata, 2) = "go" Then
            cdata = ""
            If Len(actmove) < 5 Then
                Print #2, "PLAY"
                'Print #3, "PLAY",
            Else
                Print #2, "MOVE "; actmove
                'Print #3, "MOVE "; actmove,
            End If
            Do
                dat = Input(1, #2)
                If Asc(dat) > 31 And dat <> ">" Then cdata = cdata + dat
            Loop Until dat = Chr(&H17)
            'Print #3, cdata
            oldpv = "" : actmove = ""
            Do
                bestmove = "" : cdata = ""
                ' After starting calculation with PLAY or MOVE the board is given 750 ms for the calculation
                ' Then the actual line is received with a SENDINFO command and send to Arena as 'info'
                ' If the line ends with a '.' the calculation is finished and the best move is returned with 'bestmove'
                Sleep 750, 1
                ' Flush COM buffer
                While LOC(2) > 0
                    dat = Input(LOC(2), #2)
                Wend
                Print #2, "SENDINFO"
                Do
                    dat = Input(1, #2)
                    If Asc(dat) > 31 And dat <> ">" Then cdata = cdata + dat
                Loop Until dat = Chr(&H17)
                'Print #3, cdata
                If Left(cdata, 8) = "Sendinfo" Then
                    dat = Mid(cdata, 29)
                    bestmove = Left(dat, Instr(dat, " ") - 1)
                    If Len(bestmove) < 5 Then bestmove = ""
                    If bestmove = "" And Right(dat, 1) = "." Then bestmove = Left(dat, Instr(dat, ".") - 1)
                    ' Actual line is stored in pv
                    If Right(dat, 1) = "." Then
                        pv = Left(dat, Len(dat) - 1)
                    Else
                        pv = dat
                    End If
                    ' If actual line differs from the previous, it is send to Arena as 'info'
                    If pv <> oldpv Then
                        Print "info depth "; Trim(Mid(cdata, 18, 2)); " score cp "; Trim(Mid(cdata, 21, 6)); " pv "; pv
                        oldpv = pv
                    End If
                End If
            Loop Until Right(cdata, 1) = "." And Len(bestmove) > 4
            cdata = ""
            ' 'bestmove' found is set in the board, this is not done automatically if BOARD is OFF
            Print #2, "MOVE "; bestmove
            'Print #3, "Int. MOVE "; bestmove,
            Do
                dat = Input(1, #2)
                If Asc(dat) > 31 And dat <> ">" Then cdata = cdata + dat
            Loop Until dat = Chr(&H17)
            'Print #3, cdata
            bestmove = Left(bestmove, 2) + Mid(bestmove, 4)
            Print "bestmove "; bestmove
        End If

' Action on Arena command 'quit'
Loop Until indata = "quit"

' Close open files and release COM port
Print #2, "CLOSE"
Close
End

Sub setuppos(ByRef fenstr As String)
    Dim As String tmp, b, col2move, piece, dat
    Dim As Short i, j, x, k
    i = InStr(1, fenstr, " ")
    If i = 0 Then
        Print "fen error"
        Exit Sub    ' wrong fen
    Else
        b = Left(fenstr, i - 1)	  ' positions
        k = i + 1
    End If
    i = InStr(k, fenstr, " ")
    If i = 0 Then
        Print "fen error"
        Exit Sub
    Else
        tmp = Mid(fenstr, i - 1, 1)	  ' color to move
        If tmp <> "w" And tmp <> "b" Then Print "fen error" : Exit Sub
        If tmp = "w" Then col2move = "WHITE "
        If tmp = "b" Then  col2move = "BLACK "
    End If

    ' switch to setup mode and clear board
    Print #2, "SETUP"
    Do
        dat = Input(1, #2)
    Loop Until dat = Chr(&H17)
    Print #2, "FUNCTION NEW"
    Do
        dat = Input(1, #2)
    Loop Until dat = Chr(&H17)

    ' convert fen postions to OSA positions and send each row
    k = 1 : b = b + "/"
    FOR i = 8 TO 1 Step -1      ' all rows
        x = 1                   ' set A-column
        tmp = ""
        Do
            If Asc(Mid(b, k, 1)) > 65 And Asc(Mid(b, k, 1)) < 83 Then	' big letter (white)
               Select Case Mid(b, k, 1)
                Case "P"
                    piece = "WP"
                Case "K"
                    piece = "WK"
                Case "Q"
                    piece = "WQ"
                Case "N"
                    piece = "WN"
                Case "B"
                    piece = "WB"
                Case "R"
                    piece = "WR"
               End Select
               tmp = tmp + piece + Chr(96 + x) + Chr(48 + i) + ","
               x = x + 1
            End If
            If Asc(Mid(b, k, 1)) > 97 And Asc(Mid(b, k, 1)) < 115 Then	' small letter (black)
               Select Case Mid(b, k, 1)
                Case "p"
                    piece = "BP"
                Case "k"
                    piece = "BK"
                Case "q"
                    piece = "BQ"
                Case "n"
                    piece = "BN"
                Case "b"
                    piece = "BB"
                Case "r"
                    piece = "BR"
               End Select
               tmp = tmp + piece + Chr(96 + x) + Chr(48 + i) + ","
               x = x + 1
            End If
            If Asc(Mid(b, k, 1)) > 48 And Asc(Mid(b, k, 1)) < 57 Then	' digits (empty fields)
               For j = 1 To ValUInt(Mid(b, k, 1))
                    x = x + 1
               Next j
            End If 
            k = k + 1
        Loop Until Mid(b, k, 1) = "/"
        If Right(tmp, 1) = "," Then tmp = Left(tmp, Len(tmp) - 1)
        If tmp <> "" Then
            ' Flush COM buffer
            While LOC(2) > 0
                dat = Input(LOC(2), #2)
            Wend
            Print #2, "POSITION "; col2move; tmp
            Do
                dat = Input(1, #2)
            Loop Until dat = Chr(&H17)
        End If
        k = k + 1
    NEXT i

    ' switch back to normal mode
    Print #2, "BOARD OFF"
    Do
        dat = Input(1, #2)
    Loop Until dat = Chr(&H17)
    Print #2, "NORMAL"
    Do
        dat = Input(1, #2)
    Loop Until dat = Chr(&H17)
End Sub
